use crate::error::ResponseError;
use crate::models::user::{User, UserRole};
use crate::result::ResponseResult;
use crate::services::user_service::UserService;
use rand::distributions::Alphanumeric;
use rand::Rng;
use rand_chacha::rand_core::SeedableRng;
use rand_chacha::ChaCha20Rng;
use sha2::{Digest, Sha256};
use std::future::Future;

pub type AuthSession<Pool> = axum_login::AuthSession<UserService<Pool>>;

pub async fn authenticate<'a, Ok, Fut>(
    maybe_user: Option<&'a User>,
    roles: impl IntoIterator<Item = UserRole>,
    authenticated: impl FnOnce(&'a User) -> Fut,
) -> ResponseResult<Ok>
where
    Fut: Future<Output = ResponseResult<Ok>>,
{
    if let Some(user) = maybe_user {
        let int_roles = roles
            .into_iter()
            .map(UserRole::into_i16)
            .collect::<Vec<_>>();

        if int_roles.contains(&user.role) {
            authenticated(user).await
        } else {
            Err(ResponseError::Forbidden)
        }
    } else {
        Err(ResponseError::Unauthorized)
    }
}

pub(crate) fn hash_password(password: &str, salt: &[u8], output: &mut [u8; 32]) {
    let hasher = Sha256::new()
        .chain_update(password)
        .chain_update("$")
        .chain_update(salt);
    output.copy_from_slice(hasher.finalize().as_slice());
}

pub(crate) fn generate_salt() -> Vec<u8> {
    ChaCha20Rng::from_entropy()
        .sample_iter(Alphanumeric)
        .take(16)
        .collect()
}

#[cfg(test)]
mod tests {
    use super::*;
    use crate::ok;
    use crate::result::Success;

    #[tokio::test]
    async fn test_authenticate_user_none() {
        let user = None;
        assert_eq!(
            authenticate(user, [UserRole::Admin], |_| async { ok!() }).await,
            Err(ResponseError::Unauthorized)
        );
    }

    #[tokio::test]
    async fn test_authenticate_user_unauthorized() {
        let user = Some(User {
            email: "test@test.com".to_string(),
            role: UserRole::Guest.into_i16(),
            hash: String::default(),
            salt: String::default(),
        });
        assert_eq!(
            authenticate(user.as_ref(), [UserRole::Admin], |_| async { ok!() }).await,
            Err(ResponseError::Forbidden)
        );
    }

    #[tokio::test]
    async fn test_authenticate_user_authorized() {
        let user = Some(User {
            email: "test@test.com".to_string(),
            role: UserRole::Admin.into_i16(),
            hash: String::default(),
            salt: String::default(),
        });
        assert_eq!(
            authenticate(user.as_ref(), [UserRole::Admin], |_| async { ok!() }).await,
            Ok(Success::Ok(()))
        );
    }

    #[test]
    fn test_hash_password() {
        let salt = generate_salt();
        let mut hashed = [0; 32];
        hash_password("password", &salt, &mut hashed);
    }
}
