use crate::services::reservation_service::ReservationError;
use axum::http::StatusCode;
use axum::response::{IntoResponse, Response};
use axum_login::AuthnBackend;
use derive_more::Constructor;
use diesel::result::DatabaseErrorKind;
use diesel_async::pooled_connection::deadpool;
use lib::diesel::get_connection::GetConnectionError;
use lib::diesel_crud_trait::CrudError;
use lib::into_response_derive::IntoResponse;
use serde::{Deserialize, Serialize};
use thiserror::Error;

type DieselError = diesel::result::Error;

#[derive(Error, Debug)]
pub enum AppError {
    #[error(transparent)]
    PoolError(#[from] deadpool::PoolError),
    #[error(transparent)]
    CrudError(CrudError),
    #[error(transparent)]
    DatabaseError(#[from] diesel::result::Error),
    #[error("Auth session error: {0}")]
    AuthSessionError(String),
    #[error("Bad request: {0}")]
    BadRequest(String),
    #[error("Resource not found")]
    NotFound,
    #[error("Unauthorized. Please log in at /auth/login")]
    Unauthorized,
    #[error("Forbidden. Insufficient permissions")]
    Forbidden,
    #[error("Internal server error: {0}")]
    InternalServerError(String),
    #[error(transparent)]
    JsonError(#[from] serde_json::Error),
    #[error(transparent)]
    TimeParseError(#[from] chrono::ParseError),
    #[error(transparent)]
    IdParseError(#[from] std::num::ParseIntError),
    #[error(transparent)]
    FromUtf8Error(#[from] std::string::FromUtf8Error),
    #[error("Base64 error: {0}")]
    Base64Error(String),
}

#[derive(Debug, Error, PartialEq)]
pub enum ResponseError {
    #[error("{0}")]
    BadRequest(String), // 400
    #[error("User is not authenticated. Please log in at /auth/login")]
    Unauthorized, // 401
    #[error("Forbidden. Insufficient permissions")]
    Forbidden, // 403
    #[error("{0}")]
    NotFound(String), // 404
    #[error("{0}")]
    Conflict(String), // 409
    #[error("{0}")]
    InternalServerError(String), // 500
}

#[derive(Debug, Serialize, Deserialize, IntoResponse, Constructor)]
pub(crate) struct ErrorResponseBody {
    pub kind: i32,
    pub error: String,
}

impl IntoResponse for ResponseError {
    fn into_response(self) -> Response {
        match self {
            ResponseError::BadRequest(error) => {
                (StatusCode::BAD_REQUEST, ErrorResponseBody::new(400, error))
            }
            ResponseError::Unauthorized => (
                StatusCode::UNAUTHORIZED,
                ErrorResponseBody::new(401, self.to_string()),
            ),
            ResponseError::Forbidden => (
                StatusCode::FORBIDDEN,
                ErrorResponseBody::new(403, self.to_string()),
            ),
            ResponseError::NotFound(error) => {
                (StatusCode::NOT_FOUND, ErrorResponseBody::new(404, error))
            }
            ResponseError::Conflict(error) => {
                (StatusCode::CONFLICT, ErrorResponseBody::new(409, error))
            }
            ResponseError::InternalServerError(error) => (
                StatusCode::INTERNAL_SERVER_ERROR,
                ErrorResponseBody::new(500, error),
            ),
        }
        .into_response()
    }
}

impl From<GetConnectionError> for ResponseError {
    fn from(value: GetConnectionError) -> Self {
        match value {
            GetConnectionError::PoolError(error) => Self::InternalServerError(error.to_string()),
            GetConnectionError::DieselError(error) => Self::InternalServerError(error.to_string()),
        }
    }
}

impl From<CrudError> for ResponseError {
    fn from(value: CrudError) -> Self {
        match &value {
            CrudError::NotFound => Self::NotFound(value.to_string()),
            CrudError::Other(diesel::result::Error::DatabaseError(
                DatabaseErrorKind::UniqueViolation,
                error,
            )) => Self::Conflict(error.message().to_string()),
            _ => Self::InternalServerError(value.to_string()),
        }
    }
}

impl From<ReservationError> for ResponseError {
    fn from(value: ReservationError) -> Self {
        match value {
            ReservationError::NotFound => Self::NotFound(value.to_string()),
            ReservationError::Other(error) => Self::InternalServerError(error),
            _ => Self::BadRequest(value.to_string()),
        }
    }
}

impl From<deadpool::PoolError> for ResponseError {
    fn from(value: deadpool::PoolError) -> Self {
        Self::InternalServerError(value.to_string())
    }
}

impl<T: AuthnBackend> From<axum_login::Error<T>> for ResponseError {
    fn from(value: axum_login::Error<T>) -> Self {
        Self::BadRequest(value.to_string())
    }
}

impl From<AppError> for ResponseError {
    fn from(value: AppError) -> Self {
        match value {
            AppError::NotFound => Self::NotFound(value.to_string()),
            AppError::BadRequest(error) => Self::BadRequest(error),
            AppError::Unauthorized => Self::Unauthorized,
            AppError::Forbidden => Self::Forbidden,
            AppError::CrudError(error) => error.into(),
            _ => Self::InternalServerError(value.to_string()),
        }
    }
}

impl IntoResponse for AppError {
    fn into_response(self) -> Response {
        match self {
            Self::NotFound => (StatusCode::NOT_FOUND, self.to_string()).into_response(),
            Self::BadRequest(_) => (StatusCode::BAD_REQUEST, self.to_string()).into_response(),
            Self::Unauthorized => (StatusCode::UNAUTHORIZED, self.to_string()).into_response(),
            Self::Forbidden => (StatusCode::FORBIDDEN, self.to_string()).into_response(),
            otherwise => (StatusCode::INTERNAL_SERVER_ERROR, otherwise.to_string()).into_response(),
        }
    }
}

impl From<GetConnectionError> for AppError {
    fn from(value: GetConnectionError) -> Self {
        match value {
            GetConnectionError::PoolError(error) => Self::PoolError(error),
            GetConnectionError::DieselError(error) => Self::DatabaseError(error),
        }
    }
}

impl From<base64ct::Error> for AppError {
    fn from(value: base64ct::Error) -> Self {
        Self::Base64Error(value.to_string())
    }
}

impl From<CrudError> for AppError {
    fn from(value: CrudError) -> Self {
        match value {
            CrudError::NotFound => Self::NotFound,
            CrudError::Other(DieselError::DatabaseError(
                DatabaseErrorKind::ForeignKeyViolation,
                error,
            )) => Self::BadRequest(format!(
                "Foreign key violation: '{}' on constraint '{}'",
                error.message(),
                error.constraint_name().unwrap_or_default()
            )),
            otherwise => Self::CrudError(otherwise),
        }
    }
}
