#[cfg(test)]
#[macro_use]
extern crate rstest;

use crate::database::create_pool;
use crate::services::session_service::SessionService;
use crate::services::user_service::UserService;
use axum::middleware::{from_fn, Next};
use axum::response::IntoResponse;
use axum::Router;
use axum_login::tower_sessions::SessionManagerLayer;
use axum_login::AuthManagerLayerBuilder;
use lib::axum::app::AppBuilder;
use lib::diesel::get_connection::GetConnection;
use tower_sessions::cookie::time::Duration;
use tower_sessions::Expiry;

mod auth;
mod config;
mod database;
mod error;
mod models;
mod result;
mod routes;
mod schema;
mod services;
#[cfg(test)]
mod test;

pub fn create_app<Pool>(pool: Pool) -> Router
where
    Pool: GetConnection + 'static,
{
    let session_service = SessionService::new(pool.clone());
    let user_service = UserService::new(pool.clone());

    let session_layer = SessionManagerLayer::new(session_service)
        .with_secure(false)
        .with_expiry(Expiry::OnInactivity(Duration::days(
            config::SESSION_MAX_AGE,
        )));
    let auth_layer = AuthManagerLayerBuilder::new(user_service.clone(), session_layer).build();
    AppBuilder::new()
        .routes([
            routes::router(),
            routes::hotel::router()
                .with_state(pool.clone())
                .login_required::<Pool>(),
            routes::reservation::router()
                .with_state(pool.clone())
                .login_required::<Pool>(),
            routes::room::router()
                .with_state(pool.clone())
                .login_required::<Pool>(),
            routes::task::router()
                .with_state(pool.clone())
                .login_required::<Pool>(),
            routes::user::router()
                .with_state(pool)
                .login_required::<Pool>(),
            routes::auth::router().with_state(user_service),
        ])
        .layer(auth_layer)
        .build()
}

#[tokio::main]
async fn main() {
    let pool = create_pool().unwrap();
    AppBuilder::from_router(create_app(pool))
        .serve()
        .await
        .unwrap();
}

trait LoginRequired {
    fn login_required<Pool>(self) -> Self
    where
        Pool: GetConnection + 'static;
}

impl<S> LoginRequired for Router<S>
where
    S: Clone + Send + Sync + 'static,
{
    fn login_required<Pool>(self) -> Self
    where
        Pool: GetConnection + Send + Sync + 'static,
    {
        self.route_layer(from_fn(
            |auth_session: axum_login::AuthSession<UserService<Pool>>, req, next: Next| async move {
                if auth_session.user.is_some() {
                    next.run(req).await
                } else {
                    axum::http::StatusCode::UNAUTHORIZED.into_response()
                }
            },
        ))
    }
}
