use chrono::{NaiveDateTime, Utc};
use derive_more::Constructor;
use diesel::{AsChangeset, Insertable, Queryable, Selectable};
use lib::diesel_crud_derive::DieselCrud;
use lib::into_response_derive::IntoResponse;
use serde::{Deserialize, Serialize};
use std::borrow::Cow;
use validator::{Validate, ValidationError};

#[derive(Clone, Validate, Insertable, Deserialize, Constructor)]
#[validate(schema(function = "validate_create_reservation"))]
#[diesel(table_name = crate::schema::reservation)]
pub struct CreateReservation {
    pub room_id: i32,
    pub start: NaiveDateTime,
    pub end: NaiveDateTime,
    #[validate(email)]
    pub user: String,
}

fn validate_create_reservation(reservation: &CreateReservation) -> Result<(), ValidationError> {
    _validate_reservation(&reservation.start, &reservation.end)
}

#[derive(
    Clone,
    Validate,
    Queryable,
    Selectable,
    AsChangeset,
    Serialize,
    Deserialize,
    IntoResponse,
    DieselCrud,
)]
#[diesel_crud(insert = CreateReservation)]
#[validate(schema(function = "validate_reservation"))]
#[diesel(table_name = crate::schema::reservation)]
#[diesel(check_for_backend(diesel::pg::Pg))]
pub struct Reservation {
    #[diesel_crud(pk)]
    pub id: i32,
    pub room_id: i32,
    pub start: NaiveDateTime,
    pub end: NaiveDateTime,
    #[validate(email)]
    pub user: String,
    pub checked_in: bool,
}

fn validate_reservation(reservation: &Reservation) -> Result<(), ValidationError> {
    _validate_reservation(&reservation.start, &reservation.end)
}

fn _validate_reservation(
    start: &NaiveDateTime,
    end: &NaiveDateTime,
) -> Result<(), ValidationError> {
    if start >= end {
        let mut err = ValidationError::new("start must be before end");
        err.add_param(Cow::from("start"), start);
        err.add_param(Cow::from("end"), end);
        Err(err)
    } else if start < &Utc::now().naive_utc() {
        let mut err = ValidationError::new("start must be in the future");
        err.add_param(Cow::from("start"), start);
        Err(err)
    } else {
        Ok(())
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn test_validate_reservation_ok() {
        let reservation = CreateReservation {
            room_id: 1,
            start: Utc::now().naive_utc() + chrono::Duration::days(1),
            end: Utc::now().naive_utc() + chrono::Duration::days(2),
            user: "a@example.com".to_string(),
        };
        assert!(validate_create_reservation(&reservation).is_ok());
    }

    #[test]
    fn test_validate_reservation_start_after_end() {
        let reservation = CreateReservation {
            room_id: 1,
            start: Utc::now().naive_utc() + chrono::Duration::days(1),
            end: Utc::now().naive_utc(),
            user: "a@example.com".to_string(),
        };
        assert!(validate_create_reservation(&reservation).is_err());
    }

    #[test]
    fn test_validate_reservation_start_in_past() {
        let reservation = CreateReservation {
            room_id: 1,
            start: Utc::now().naive_utc() - chrono::Duration::days(1),
            end: Utc::now().naive_utc(),
            user: "a@example.com".to_string(),
        };
        assert!(validate_create_reservation(&reservation).is_err());
    }
}
