use crate::error::AppError;
use chrono::{DateTime, NaiveDateTime};
use diesel::{Insertable, Queryable, Selectable};
use lib::diesel_crud_derive::{DieselCrudCreate, DieselCrudDelete, DieselCrudRead};
use serde_json::Value;
use std::collections::HashMap;
use std::time::SystemTime;
use tower_sessions::cookie::time::OffsetDateTime;
use tower_sessions::session;
use tower_sessions::session::Record;

#[derive(Insertable, Queryable, Selectable, DieselCrudCreate, DieselCrudRead, DieselCrudDelete)]
#[diesel(table_name = crate::schema::session)]
#[diesel(check_for_backend(diesel::pg::Pg))]
pub struct Session {
    #[diesel_crud(pk)]
    pub id: String,
    pub data: Value,
    pub expiry_date: NaiveDateTime,
}

impl TryFrom<Record> for Session {
    type Error = AppError;
    fn try_from(value: Record) -> Result<Self, Self::Error> {
        Ok(Self {
            id: value.id.0.to_string(),
            data: serde_json::to_value(value.data)?,
            expiry_date: DateTime::from_timestamp_micros(value.expiry_date.unix_timestamp())
                .ok_or(AppError::InternalServerError(
                    "Failed to convert time".into(),
                ))?
                .naive_utc(),
        })
    }
}

impl Session {
    pub fn try_into_record(self) -> Result<Record, AppError> {
        Ok(Record {
            id: session::Id(self.id.trim().parse::<i128>()?),
            expiry_date: self.expiry_date.into_offset_date_time(),
            data: self.data.into_hash_map(),
        })
    }
}

trait IntoOffsetDateTime {
    fn into_offset_date_time(self) -> OffsetDateTime;
}

impl IntoOffsetDateTime for NaiveDateTime {
    fn into_offset_date_time(self) -> OffsetDateTime {
        let system_time = SystemTime::from(self.and_utc());
        OffsetDateTime::from(system_time)
    }
}

impl IntoHashMap<String, Value> for Value {
    fn into_hash_map(self) -> HashMap<String, Value> {
        if let Value::Object(map) = self {
            map.keys().cloned().zip(map.values().cloned()).collect()
        } else {
            HashMap::new()
        }
    }
}

trait IntoHashMap<Key, Value> {
    fn into_hash_map(self) -> HashMap<Key, Value>;
}

#[cfg(test)]
mod tests {
    // TODO
}
