use diesel::{AsChangeset, Insertable, Queryable, Selectable};
use lib::diesel_crud_derive::DieselCrud;
use lib::into_response_derive::IntoResponse;
use serde::{Deserialize, Serialize};
use std::borrow::Cow;
use strum::Display;
use validator::{Validate, ValidationError};

#[derive(
    Queryable, Selectable, AsChangeset, Validate, Serialize, Deserialize, IntoResponse, DieselCrud,
)]
#[diesel_crud(insert = CreateTask)]
#[diesel(table_name = crate::schema::task)]
#[diesel(check_for_backend(diesel::pg::Pg))]
pub struct Task {
    #[diesel_crud(pk)]
    pub id: i32,
    #[validate(range(min = 1))]
    pub room_id: i32,
    #[validate(length(min = 1))]
    pub description: String,
    #[validate(custom(function = "validate_task_status"))]
    pub status: String, // TODO use Enum
}

#[derive(Insertable, Validate, Deserialize)]
#[diesel(table_name = crate::schema::task)]
#[diesel(check_for_backend(diesel::pg::Pg))]
pub struct CreateTask {
    #[validate(range(min = 1))]
    pub room_id: i32,
    #[validate(length(min = 1))]
    pub description: String,
    #[validate(custom(function = "validate_task_status"))]
    pub status: String,
}

fn validate_task_status(status: &str) -> Result<(), ValidationError> {
    if ["todo", "in_progress", "done"].contains(&status) {
        Ok(())
    } else {
        let mut err = ValidationError::new("status must be 'todo', 'in_progress' or 'done'");
        err.add_param(Cow::from("status"), &status);
        Err(err)
    }
}

#[derive(Debug, Clone, Copy, Display, Deserialize)]
#[strum(serialize_all = "snake_case")]
pub enum TaskStatus {
    Todo,
    InProgress,
    Done,
}

#[cfg(test)]
mod tests {
    use super::*;

    #[rstest]
    fn test_validate_task_status_legal_status(
        #[values("todo", "in_progress", "done")] status: &str,
    ) {
        assert!(validate_task_status(status).is_ok());
    }

    #[rstest]
    fn test_validate_task_status_illegal_status(#[values("illegal", "status")] status: &str) {
        assert!(validate_task_status(status).is_err());
    }

    #[test]
    fn test_enum_display() {
        assert_eq!(TaskStatus::Todo.to_string(), "todo");
        assert_eq!(TaskStatus::InProgress.to_string(), "in_progress");
        assert_eq!(TaskStatus::Done.to_string(), "done");
    }
}
