use crate::auth::{generate_salt, hash_password};
use crate::error::AppError;
use axum_login::AuthUser;
use base64ct::{Base64, Encoding};
use derive_more::From;
use diesel::{Insertable, Queryable, Selectable};
use lib::diesel_crud_derive::{DieselCrudCreate, DieselCrudDelete, DieselCrudList, DieselCrudRead};
use lib::into_response_derive::IntoResponse;
use secrecy::{ExposeSecret, SecretString};
use serde::{Deserialize, Serialize};
use strum::EnumIs;
use validator::Validate;

#[derive(
    Debug,
    Clone,
    Queryable,
    Selectable,
    Insertable,
    Validate,
    Serialize,
    Deserialize,
    IntoResponse,
    DieselCrudRead,
    DieselCrudCreate,
    DieselCrudDelete,
    DieselCrudList,
)]
#[diesel(table_name = crate::schema::user)]
#[diesel(check_for_backend(diesel::pg::Pg))]
pub struct User {
    #[diesel_crud(pk)]
    #[validate(email)]
    pub email: String,
    pub hash: String,
    pub salt: String,
    pub role: i16, // TODO: use enum
}

#[derive(Validate, Deserialize)]
pub struct UserCredentials {
    #[validate(email)]
    pub email: String,
    pub password: SecretString,
}

#[derive(Clone, Validate, Deserialize)]
pub struct CreateUser {
    #[validate(email)]
    pub email: String,
    pub password: SecretString,
    pub role: Option<UserRole>,
}

#[derive(Debug, Clone, Copy, PartialEq, Eq, Default, Serialize, Deserialize, EnumIs)]
#[serde(rename_all = "SCREAMING_SNAKE_CASE")]
pub enum UserRole {
    Admin,
    Receptionist,
    Janitor,
    #[default]
    Guest,
}

impl CreateUser {
    pub fn new(email: impl Into<String>, password: impl Into<String>) -> Self {
        Self {
            email: email.into(),
            password: SecretString::new(password.into()),
            role: None,
        }
    }
}

impl User {
    pub fn hash_from_credentials(create_user: CreateUser) -> Result<Self, AppError> {
        let salt = generate_salt();
        let mut hashed = [0; 32];
        hash_password(create_user.password.expose_secret(), &salt, &mut hashed);

        Ok(Self {
            email: create_user.email.clone(),
            hash: Base64::encode_string(&hashed),
            salt: String::from_utf8(salt.to_vec())?,
            role: create_user.role.unwrap_or_default() as i16,
        })
    }
}

impl UserRole {
    pub fn into_i16(self) -> i16 {
        self as i16
    }
}

impl AuthUser for User {
    type Id = String;

    fn id(&self) -> Self::Id {
        self.email.clone()
    }

    fn session_auth_hash(&self) -> &[u8] {
        // TODO use decode?
        self.hash.as_bytes()
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn test_hash_from_credentials() {
        let credentials = CreateUser {
            email: "me@test.com".to_string(),
            password: SecretString::new("password".to_string()),
            role: None,
        };
        let user = User::hash_from_credentials(credentials.clone()).unwrap();
        assert_eq!(user.email, credentials.email);
        assert_eq!(user.role, UserRole::Guest.into_i16());
        assert!(!user.hash.is_empty());
        assert!(!user.salt.is_empty());
    }
}
