use crate::error::ResponseError;
use axum::http::StatusCode;
use axum::response::{IntoResponse, Response};

pub type ResponseResult<T = ()> = Result<Success<T>, ResponseError>;

#[derive(Debug, Clone, PartialEq)]
pub enum Success<T = ()> {
    Ok(T),
    Created(T),
}

impl<T: IntoResponse> IntoResponse for Success<T> {
    fn into_response(self) -> Response {
        match self {
            Success::Ok(value) => (StatusCode::OK, value).into_response(),
            Success::Created(value) => (StatusCode::CREATED, value).into_response(),
        }
    }
}

impl<T> From<T> for Success<T> {
    fn from(value: T) -> Self {
        Success::Ok(value)
    }
}

#[macro_export]
macro_rules! ok {
    () => {
        Ok($crate::result::Success::Ok(()))
    };
    ($expr:expr) => {
        Ok($crate::result::Success::Ok($expr))
    };
}

#[macro_export]
macro_rules! created {
    () => {
        Ok($crate::result::Success::Created(()))
    };
    ($expr:expr) => {
        Ok($crate::result::Success::Created($expr))
    };
}

#[macro_export]
macro_rules! bad_request {
    ($expr:expr) => {
        Err($crate::error::ResponseError::BadRequest($expr.to_string()))
    };
}

#[macro_export]
macro_rules! not_found {
    ($expr:expr) => {
        Err($crate::error::ResponseError::NotFound($expr.to_string()))
    };
}

#[macro_export]
macro_rules! unauthorized {
    ($expr:expr) => {
        Err($crate::error::ResponseError::Unauthorized(
            $expr.to_string(),
        ))
    };
}

#[macro_export]
macro_rules! forbidden {
    () => {
        Err($crate::error::ResponseError::Forbidden)
    };
}

#[macro_export]
macro_rules! internal_server_error {
    ($expr:expr) => {
        Err($crate::error::ResponseError::InternalServerError(
            $expr.to_string(),
        ))
    };
}
