use crate::auth::AuthSession;
use crate::error::ResponseError;
use crate::models::user::{CreateUser, User, UserCredentials};
use crate::result::{ResponseResult, Success};
use crate::services::user_service::UserService;
use crate::{bad_request, internal_server_error, ok};
use axum::extract::State;
use axum::Json;
use axum_valid::Valid;
use lib::diesel::get_connection::GetConnection;
// router!(
//     "/auth",
//     routes!(
//         post "/login" => login::<Pool>,
//         post "/register" => register::<Pool>
//     ),
//     Pool: GetConnection -> UserService
// );

pub fn router<Pool: GetConnection + 'static>() -> axum::Router<UserService<Pool>> {
    axum::Router::new().nest(
        "/auth",
        axum::Router::new()
            .route("/login", axum::routing::post(login::<Pool>))
            .route("/register", axum::routing::post(register::<Pool>)),
    )
}

async fn login<Pool>(
    mut auth_session: AuthSession<Pool>,
    Valid(Json(credentials)): Valid<Json<UserCredentials>>,
) -> ResponseResult
where
    Pool: GetConnection,
{
    let user = match auth_session.authenticate(credentials).await {
        Ok(Some(user)) => user,
        Ok(None) => return bad_request!("Invalid credentials"),
        Err(error) => return internal_server_error!(error),
    };
    auth_session.login(&user).await?;
    ok!()
}

async fn register<Pool>(
    State(service): State<UserService<Pool>>,
    Valid(Json(create_user)): Valid<Json<CreateUser>>,
) -> ResponseResult<User>
where
    Pool: GetConnection,
{
    service
        .insert(create_user)
        .await
        .map(Success::Created)
        .map_err(ResponseError::from)
}

#[cfg(test)]
mod tests {
    use crate::create_app;
    use crate::error::ErrorResponseBody;
    use crate::models::user::{CreateUser, User, UserRole};
    use crate::test::create_test_containers_pool;
    use axum::http::request::Builder;
    use axum::http::{Request, StatusCode};
    use axum::Router;
    use futures::executor::block_on;
    use lib::axum::traits::BuildJson;
    use lib::serde::traits::DeserializeInto;
    use secrecy::ExposeSecret;
    use serde::ser::SerializeStruct;
    use serde::{Serialize, Serializer};
    use testcontainers_modules::postgres::Postgres;
    use testcontainers_modules::testcontainers::ContainerAsync;
    use tower::ServiceExt;

    fn register() -> Builder {
        Request::builder().uri("/auth/register").method("POST")
    }

    #[fixture]
    #[once]
    fn setup() -> Setup {
        block_on(async {
            let test_container = create_test_containers_pool().await.unwrap();
            let app = create_app(test_container.pool);
            Setup {
                _container: test_container.container,
                app,
            }
        })
    }

    struct Setup {
        _container: ContainerAsync<Postgres>,
        app: Router,
    }

    #[rstest]
    #[tokio::test(flavor = "multi_thread")]
    async fn test_register_created(setup: &Setup) {
        let create_user = CreateUser::new("test_register_created@email.com", "password");
        let create_email = create_user.email.clone();

        let response = setup
            .app
            .clone()
            .oneshot(register().json(create_user).unwrap())
            .await
            .unwrap();

        let status = response.status();
        let user: User = response.deserialize_into().await.unwrap();

        assert_eq!(status, StatusCode::CREATED);
        assert_eq!(user.email, create_email);
        assert_eq!(user.role, UserRole::default() as i16);
        assert!(!user.hash.is_empty());
        assert!(!user.salt.is_empty());
    }

    #[rstest]
    #[tokio::test(flavor = "multi_thread")]
    async fn test_register_email_already_registered(setup: &Setup) {
        let create_user = CreateUser::new(
            "test_register_email_already_registered@email.com",
            "password",
        );

        let call = || async {
            setup
                .app
                .clone()
                .oneshot(register().json(create_user.clone()).unwrap())
                .await
                .unwrap()
        };
        let created = call().await;
        let conflicted = call().await;

        let conflicted_status = conflicted.status();

        let response: ErrorResponseBody = conflicted.deserialize_into().await.unwrap();

        assert_eq!(created.status(), StatusCode::CREATED);
        assert_eq!(conflicted_status, StatusCode::CONFLICT);
        assert!(
            response.error.contains("user_pkey"),
            "Error: {}",
            response.error
        );
    }

    impl Serialize for CreateUser {
        fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
        where
            S: Serializer,
        {
            let mut s = serializer.serialize_struct("CreateUser", 3)?;
            s.serialize_field("email", &self.email)?;
            s.serialize_field("password", &self.password.expose_secret())?;
            s.serialize_field("role", &self.role)?;
            s.end()
        }
    }
}
