use crate::auth::{authenticate, AuthSession};
use crate::error::ResponseError;
use crate::models::reservation::{CreateReservation, Reservation};
use crate::models::user::UserRole;
use crate::result::{ResponseResult, Success};
use crate::{forbidden, ok, GetConnection};
use axum::extract::{Path, State};
use axum::Json;
use axum_valid::Valid;
use lib::axum::wrappers::Array;
use lib::diesel_crud_trait::{
    DieselCrudCreate, DieselCrudDelete, DieselCrudList, DieselCrudRead, DieselCrudUpdate,
};
use lib::time::DateTimeInterval;
use lib::{router, routes};

router!(
    "/reservation",
    routes! {
        get "/:id" => get_reservation::<T>,
        get "/" => list_reservations::<T>,
        post "/" => create_reservation::<T>,
        patch "/" => update_reservation::<T>,
        delete "/:id" => delete_reservation::<T>,
        patch "/:id/check-in" => check_in::<T>,
        patch "/:id/check-out" => check_out::<T>,
    },
    T: Clone, GetConnection
);

async fn get_reservation<Pool>(
    auth_session: AuthSession<Pool>,
    State(pool): State<Pool>,
    Path(id): Path<i32>,
) -> ResponseResult<Reservation>
where
    Pool: GetConnection,
{
    authenticate(
        auth_session.user.as_ref(),
        [UserRole::Admin, UserRole::Receptionist, UserRole::Guest],
        |user| async {
            let result = Reservation::read(id, pool.get().await?.as_mut()).await?;
            if user.role == UserRole::Admin as i16
                || user.role == UserRole::Receptionist as i16
                || result.user == user.email
            {
                ok!(result)
            } else {
                forbidden!()
            }
        },
    )
    .await
}

async fn list_reservations<Pool>(
    auth_session: AuthSession<Pool>,
    State(pool): State<Pool>,
) -> ResponseResult<Array<Reservation>>
where
    Pool: GetConnection,
{
    authenticate(
        auth_session.user.as_ref(),
        [UserRole::Admin, UserRole::Receptionist],
        |_| async {
            Reservation::list(pool.get().await?.as_mut())
                .await
                .map(Array::from)
                .map(Success::Ok)
                .map_err(Into::into)
        },
    )
    .await
}

async fn create_reservation<Pool>(
    auth_session: AuthSession<Pool>,
    State(pool): State<Pool>,
    Valid(Json(reservation)): Valid<Json<CreateReservation>>,
) -> ResponseResult<Reservation>
where
    Pool: GetConnection,
{
    authenticate(
        auth_session.user.as_ref(),
        [UserRole::Admin, UserRole::Receptionist, UserRole::Guest],
        |_| async {
            let interval = DateTimeInterval::new_safe(reservation.start, reservation.end).ok_or(
                ResponseError::BadRequest("Start date must be before end date".into()),
            )?;
            let mut conn = pool.get().await?;
            Reservation::room_available(reservation.room_id, interval, &mut conn).await?;
            Reservation::insert(reservation, &mut conn)
                .await
                .map(Success::Created)
                .map_err(Into::into)
        },
    )
    .await
}

async fn update_reservation<Pool>(
    auth_session: AuthSession<Pool>,
    State(pool): State<Pool>,
    Valid(Json(reservation)): Valid<Json<Reservation>>,
) -> ResponseResult
where
    Pool: GetConnection,
{
    authenticate(
        auth_session.user.as_ref(),
        [UserRole::Admin, UserRole::Receptionist],
        |_| async {
            let interval = DateTimeInterval::new_safe(reservation.start, reservation.end).ok_or(
                ResponseError::BadRequest("Start date must be before end date".into()),
            )?;
            let mut conn = pool.get().await?;
            Reservation::room_available(reservation.room_id, interval, &mut conn).await?;
            Reservation::update(reservation, &mut conn)
                .await
                .map(|_| Success::Ok(()))
                .map_err(Into::into)
        },
    )
    .await
}

async fn delete_reservation<Pool>(
    auth_session: AuthSession<Pool>,
    State(pool): State<Pool>,
    Path(id): Path<i32>,
) -> ResponseResult
where
    Pool: GetConnection,
{
    authenticate(
        auth_session.user.as_ref(),
        [UserRole::Admin, UserRole::Receptionist],
        |_| async {
            Reservation::delete(id, pool.get().await?.as_mut())
                .await
                .map(|_| Success::Ok(()))
                .map_err(Into::into)
        },
    )
    .await
}

async fn check_in<Pool>(
    auth_session: AuthSession<Pool>,
    State(pool): State<Pool>,
    Path(id): Path<i32>,
) -> ResponseResult
where
    Pool: GetConnection,
{
    authenticate(
        auth_session.user.as_ref(),
        [UserRole::Admin, UserRole::Receptionist],
        |_| async {
            Reservation::check_in(id, pool.get().await?.as_mut())
                .await
                .map(Success::Ok)
                .map_err(Into::into)
        },
    )
    .await
}

async fn check_out<Pool>(
    auth_session: AuthSession<Pool>,
    State(pool): State<Pool>,
    Path(id): Path<i32>,
) -> ResponseResult
where
    Pool: GetConnection,
{
    authenticate(
        auth_session.user.as_ref(),
        [UserRole::Admin, UserRole::Receptionist],
        |_| async {
            Reservation::check_out(id, pool.get().await?.as_mut())
                .await
                .map(Success::Ok)
                .map_err(Into::into)
        },
    )
    .await
}
