use crate::auth::{authenticate, AuthSession};
use crate::models::room::Room;
use crate::models::user::UserRole;
use crate::result::{ResponseResult, Success};
use crate::services::room_service::RoomQuery;
use crate::{ok, GetConnection};
use axum::extract::{Path, State};
use axum::Json;
use axum_valid::Valid;
use lib::axum::wrappers::Array;
use lib::diesel_crud_trait::{
    DieselCrudCreate, DieselCrudDelete, DieselCrudList, DieselCrudRead, DieselCrudUpdate,
};
use lib::{router, routes};

router!(
    "/room",
    routes!(
        get "/:id" => get_room::<T>,
        post "/" => create_room::<T>,
        patch "/" => update_room::<T>,
        delete "/:id" => delete_room::<T>,
        get "/" => list_rooms::<T>,
        post "/query" => query_rooms::<T>
    ),
    T: Clone, GetConnection
);

async fn get_room<Pool>(
    auth_session: AuthSession<Pool>,
    State(pool): State<Pool>,
    Path(id): Path<i32>,
) -> ResponseResult<Room>
where
    Pool: GetConnection,
{
    authenticate(
        auth_session.user.as_ref(),
        [UserRole::Admin, UserRole::Receptionist],
        |_| async {
            Room::read(id, pool.get().await?.as_mut())
                .await
                .map(Success::Ok)
                .map_err(Into::into)
        },
    )
    .await
}

async fn create_room<Pool>(
    auth_session: AuthSession<Pool>,
    State(pool): State<Pool>,
    Valid(Json(room)): Valid<Json<Room>>,
) -> ResponseResult<Room>
where
    Pool: GetConnection,
{
    authenticate(
        auth_session.user.as_ref(),
        [UserRole::Admin, UserRole::Receptionist],
        |_| async {
            Room::insert(room, pool.get().await?.as_mut())
                .await
                .map(Success::Created)
                .map_err(Into::into)
        },
    )
    .await
}

async fn update_room<Pool>(
    auth_session: AuthSession<Pool>,
    State(pool): State<Pool>,
    Valid(Json(room)): Valid<Json<Room>>,
) -> ResponseResult
where
    Pool: GetConnection,
{
    authenticate(
        auth_session.user.as_ref(),
        [UserRole::Admin, UserRole::Receptionist],
        |_| async {
            Room::update(room, pool.get().await?.as_mut()).await?;
            ok!()
        },
    )
    .await
}

async fn delete_room<Pool>(
    auth_session: AuthSession<Pool>,
    State(pool): State<Pool>,
    Path(id): Path<i32>,
) -> ResponseResult
where
    Pool: GetConnection + Send + Sync + 'static,
{
    authenticate(
        auth_session.user.as_ref(),
        [UserRole::Admin, UserRole::Receptionist],
        |_| async {
            Room::delete(id, pool.get().await?.as_mut()).await?;
            ok!()
        },
    )
    .await
}

async fn list_rooms<Pool>(
    auth_session: AuthSession<Pool>,
    State(pool): State<Pool>,
) -> ResponseResult<Array<Room>>
where
    Pool: GetConnection + Send + Sync + 'static,
{
    authenticate(
        auth_session.user.as_ref(),
        [UserRole::Admin, UserRole::Receptionist],
        |_| async {
            Room::list(pool.get().await?.as_mut())
                .await
                .map(Array::from)
                .map(Success::Ok)
                .map_err(Into::into)
        },
    )
    .await
}

async fn query_rooms<Pool>(
    auth_session: AuthSession<Pool>,
    State(pool): State<Pool>,
    Json(query): Json<RoomQuery>,
) -> ResponseResult<Array<Room>>
where
    Pool: GetConnection + Send + Sync + 'static,
{
    authenticate(
        auth_session.user.as_ref(),
        [UserRole::Admin, UserRole::Receptionist],
        |_| async {
            Room::query_rooms(query, pool.get().await?.as_mut())
                .await
                .map(Array::from)
                .map(Success::Ok)
                .map_err(Into::into)
        },
    )
    .await
}
