use crate::auth::{authenticate, AuthSession};
use crate::models::task::{CreateTask, Task};
use crate::models::user::UserRole;
use crate::result::{ResponseResult, Success};
use crate::{ok, GetConnection};
use axum::extract::{Path, State};
use axum::Json;
use axum_valid::Valid;
use lib::diesel_crud_trait::{
    DieselCrudCreate, DieselCrudDelete, DieselCrudRead, DieselCrudUpdate,
};
use lib::{router, routes};

router!(
    "/task",
    routes!(
        get "/:id" => get_task::<T>,
        post "/" => create_task::<T>,
        patch "/" => update_task::<T>,
        delete "/:id" => delete_task::<T>,
        patch "/:id/description" => set_description::<T>,
        patch "/:id/status" => set_status::<T>
    ),
    T: Clone, GetConnection
);

#[derive(serde::Deserialize)]
struct Description {
    description: String,
}

#[derive(serde::Deserialize)]
struct TaskStatus {
    status: crate::models::task::TaskStatus,
}

async fn get_task<Pool>(
    auth_session: AuthSession<Pool>,
    State(pool): State<Pool>,
    Path(id): Path<i32>,
) -> ResponseResult<Task>
where
    Pool: GetConnection,
{
    authenticate(
        auth_session.user.as_ref(),
        [UserRole::Admin, UserRole::Janitor, UserRole::Receptionist],
        |_| async {
            Task::read(id, pool.get().await?.as_mut())
                .await
                .map(Success::Ok)
                .map_err(Into::into)
        },
    )
    .await
}

async fn create_task<Pool>(
    auth_session: AuthSession<Pool>,
    State(pool): State<Pool>,
    Valid(Json(create_task)): Valid<Json<CreateTask>>,
) -> ResponseResult<Task>
where
    Pool: GetConnection,
{
    authenticate(
        auth_session.user.as_ref(),
        [UserRole::Admin, UserRole::Janitor, UserRole::Receptionist],
        |_| async {
            Task::insert(create_task, pool.get().await?.as_mut())
                .await
                .map(Success::Created)
                .map_err(Into::into)
        },
    )
    .await
}

async fn update_task<Pool>(
    auth_session: AuthSession<Pool>,
    State(pool): State<Pool>,
    Valid(Json(update_task)): Valid<Json<Task>>,
) -> ResponseResult
where
    Pool: GetConnection,
{
    authenticate(
        auth_session.user.as_ref(),
        [UserRole::Admin, UserRole::Janitor, UserRole::Receptionist],
        |_| async {
            Task::update(update_task, pool.get().await?.as_mut()).await?;
            ok!()
        },
    )
    .await
}

async fn delete_task<Pool>(
    auth_session: AuthSession<Pool>,
    State(pool): State<Pool>,
    Path(id): Path<i32>,
) -> ResponseResult
where
    Pool: GetConnection,
{
    authenticate(
        auth_session.user.as_ref(),
        [UserRole::Admin, UserRole::Janitor, UserRole::Receptionist],
        |_| async {
            Task::delete(id, pool.get().await?.as_mut()).await?;
            ok!()
        },
    )
    .await
}

async fn set_description<Pool>(
    auth_session: AuthSession<Pool>,
    State(pool): State<Pool>,
    Path(id): Path<i32>,
    Json(Description { description }): Json<Description>,
) -> ResponseResult
where
    Pool: GetConnection,
{
    authenticate(
        auth_session.user.as_ref(),
        [UserRole::Admin, UserRole::Janitor, UserRole::Receptionist],
        |_| async {
            Task::set_description(id, description, pool.get().await?.as_mut()).await?;
            ok!()
        },
    )
    .await
}

async fn set_status<Pool>(
    auth_session: AuthSession<Pool>,
    State(pool): State<Pool>,
    Path(id): Path<i32>,
    Json(TaskStatus { status }): Json<TaskStatus>,
) -> ResponseResult
where
    Pool: GetConnection,
{
    authenticate(
        auth_session.user.as_ref(),
        [UserRole::Admin, UserRole::Janitor, UserRole::Receptionist],
        |_| async {
            Task::set_status(id, status, pool.get().await?.as_mut()).await?;
            ok!()
        },
    )
    .await
}
