use crate::models::reservation::Reservation;
use crate::schema::{reservation, room};
use diesel::expression_methods::ExpressionMethods;
use diesel::result::DatabaseErrorKind;
use diesel::{BoolExpressionMethods, QueryDsl};
use diesel_async::{AsyncPgConnection, RunQueryDsl};
use lib::diesel_crud_trait::CrudError;
use lib::time::DateTimeInterval;
use serde::Serialize;
use thiserror::Error;

impl Reservation {
    pub async fn check_in(id: i32, conn: &mut AsyncPgConnection) -> Result<(), ReservationError> {
        Self::update_check_in(id, true, conn).await
    }

    pub async fn check_out(id: i32, conn: &mut AsyncPgConnection) -> Result<(), ReservationError> {
        Self::update_check_in(id, false, conn).await
    }

    pub async fn room_available(
        room_number: i32,
        DateTimeInterval { start, end }: DateTimeInterval,
        conn: &mut AsyncPgConnection,
    ) -> Result<(), ReservationError> {
        let available_rooms: i64 = room::table
            .left_outer_join(reservation::table)
            .filter(room::id.eq(room_number))
            .filter(
                reservation::id
                    .is_null()
                    .or(reservation::end.le(start))
                    .or(reservation::start.ge(end)),
            )
            .count()
            .get_result(conn)
            .await?;
        if available_rooms == 0 {
            Err(ReservationError::RoomNotAvailable)
        } else {
            Ok(())
        }
    }

    async fn update_check_in(
        id: i32,
        to: bool,
        conn: &mut AsyncPgConnection,
    ) -> Result<(), ReservationError> {
        let find_query = reservation::table.find(id);

        let reservation: Reservation = find_query.get_result(conn).await?;
        if to && reservation.checked_in {
            return Err(ReservationError::AlreadyCheckedIn);
        }
        diesel::update(find_query)
            .set(reservation::checked_in.eq(to))
            .execute(conn)
            .await?;
        Ok(())
    }
}

#[derive(Debug, PartialEq, Error, Serialize)]
pub enum ReservationError {
    #[error("Reservation not found")]
    NotFound,
    #[error("Reservation already checked in")]
    AlreadyCheckedIn,
    #[error("Room does not exist")]
    RoomNotFound,
    #[error("User does not exist")]
    UserNotFound,
    #[error("Room is not available")]
    RoomNotAvailable,
    #[error("Database error: {0}")]
    Other(String),
}

impl From<CrudError> for ReservationError {
    fn from(value: CrudError) -> Self {
        match value {
            CrudError::NotFound => ReservationError::NotFound,
            CrudError::PoolError(pool_error) => ReservationError::Other(pool_error),
            CrudError::Other(diesel_error) => diesel_error.into(),
        }
    }
}

type DieselError = diesel::result::Error;

impl From<DieselError> for ReservationError {
    fn from(error: DieselError) -> Self {
        match &error {
            DieselError::NotFound => ReservationError::NotFound,
            DieselError::DatabaseError(DatabaseErrorKind::ForeignKeyViolation, info) => {
                if info.constraint_name().unwrap_or_default().contains("room") {
                    ReservationError::RoomNotFound
                } else if info.constraint_name().unwrap_or_default().contains("user") {
                    ReservationError::UserNotFound
                } else {
                    ReservationError::Other(error.to_string())
                }
            }
            _ => ReservationError::Other(error.to_string()),
        }
    }
}

#[cfg(test)]
mod tests {
    use super::*;
    use crate::models::hotel::{CreateHotel, Hotel};
    use crate::models::reservation::CreateReservation;
    use crate::models::room::Room;
    use crate::models::user::{CreateUser, User};
    use crate::schema::{hotel, room, user};
    use crate::test::create_test_containers_pool;
    use chrono::{Duration, Utc};
    use diesel::dsl::insert_into;
    use diesel_async::AsyncPgConnection;
    use futures::executor::block_on;
    use lib::diesel::pool::PgPool;
    use lib::diesel_crud_trait::{DieselCrudCreate, DieselCrudRead};
    use secrecy::SecretString;
    use testcontainers_modules::postgres::Postgres;
    use testcontainers_modules::testcontainers::ContainerAsync;

    #[rstest]
    #[tokio::test(flavor = "multi_thread")]
    async fn test_check_in(setup: &Setup) {
        let mut conn = setup.pool.get().await.unwrap();
        Reservation::check_in(setup.reservation.id, &mut conn)
            .await
            .unwrap();

        assert!(
            Reservation::read(setup.reservation.id, &mut conn)
                .await
                .unwrap()
                .checked_in
        );
    }

    #[rstest]
    #[tokio::test(flavor = "multi_thread")]
    async fn test_check_out(setup: &Setup) {
        let mut conn = setup.pool.get().await.unwrap();
        Reservation::check_out(setup.reservation.id, &mut conn)
            .await
            .unwrap();

        assert!(
            !Reservation::read(setup.reservation.id, &mut conn)
                .await
                .unwrap()
                .checked_in
        );
    }

    #[rstest]
    #[case::available(
        Duration::days(12),
        Duration::days(14),
        Ok(())
    )]
    #[case::start_and_end_in_interval(
        Duration::days(2),
        Duration::days(4),
        Err(ReservationError::RoomNotAvailable)
    )]
    #[case::start_in_interval(
        Duration::days(6),
        Duration::days(10),
        Err(ReservationError::RoomNotAvailable)
    )]
    #[case::end_in_interval(
        Duration::days(-2),
        Duration::days(2),
        Err(ReservationError::RoomNotAvailable)
    )]
    #[case::start_before_and_end_after_interval(
        Duration::days(-2),
        Duration::days(11),
        Err(ReservationError::RoomNotAvailable)
    )]
    #[tokio::test(flavor = "multi_thread")]
    async fn test_room_available(
        setup: &Setup,
        #[case] start: Duration,
        #[case] end: Duration,
        #[case] expected: Result<(), ReservationError>,
    ) {
        let mut conn = setup.pool.get().await.unwrap();

        let now = Utc::now().naive_utc();
        let start = now + start;
        let end = now + end;

        assert_eq!(
            Reservation::room_available(setup.reservation.room_id, (start, end).into(), &mut conn)
                .await,
            expected
        );
    }

    #[rstest]
    #[tokio::test(flavor = "multi_thread")]
    async fn test_room_available_no_reservations(setup: &Setup) {
        let mut conn = setup.pool.get().await.unwrap();
        let room = Room::insert(
            Room {
                id: 2,
                hotel_id: setup.hotel.id,
                beds: 1,
                size: 1,
            },
            &mut conn,
        )
        .await
        .unwrap();

        let now = Utc::now().naive_utc();
        let start = now;
        let end = now + Duration::days(10);

        assert_eq!(
            Reservation::room_available(room.id, (start, end).into(), &mut conn).await,
            Ok(())
        );
    }

    #[fixture]
    #[once]
    fn setup() -> Setup {
        block_on(async {
            let test_container = create_test_containers_pool().await.unwrap();
            let pool = test_container.pool;
            let mut conn = pool.get().await.unwrap();

            let hotel = insert_hotel(&mut conn).await;
            let user = insert_user(&mut conn).await;
            let room = insert_room(&mut conn, hotel.id).await;
            let reservation = insert_reservation(&mut conn, room.id, user.email).await;

            Setup {
                _container: test_container.container,
                pool,
                hotel,
                reservation,
            }
        })
    }

    struct Setup {
        _container: ContainerAsync<Postgres>,
        pool: PgPool,
        hotel: Hotel,
        reservation: Reservation,
    }

    async fn insert_hotel(conn: &mut AsyncPgConnection) -> Hotel {
        insert_into(hotel::table)
            .values(CreateHotel::new(
                None,
                "test".to_string(),
                "test".to_string(),
            ))
            .get_result(conn)
            .await
            .unwrap()
    }

    async fn insert_room(conn: &mut AsyncPgConnection, hotel_id: i32) -> Room {
        insert_into(room::table)
            .values(Room {
                id: 1,
                hotel_id,
                beds: 1,
                size: 1,
            })
            .get_result(conn)
            .await
            .unwrap()
    }

    async fn insert_user(conn: &mut AsyncPgConnection) -> User {
        insert_into(user::table)
            .values(
                User::hash_from_credentials(CreateUser {
                    email: "test_again@test.com".to_string(),
                    password: SecretString::new("test".to_string()),
                    role: None,
                })
                .unwrap(),
            )
            .get_result(conn)
            .await
            .unwrap()
    }

    async fn insert_reservation(
        conn: &mut AsyncPgConnection,
        room_id: i32,
        user: String,
    ) -> Reservation {
        let now = Utc::now().naive_utc();
        insert_into(reservation::table)
            .values(CreateReservation {
                room_id,
                user,
                start: now,
                end: now + Duration::days(7),
            })
            .get_result(conn)
            .await
            .unwrap()
    }
}
