use crate::error::AppError;
use crate::models::room::Room;
use crate::schema::{reservation, room};
use diesel::expression_methods::ExpressionMethods;
use diesel::{BoolExpressionMethods, QueryDsl};
use diesel_async::{AsyncPgConnection, RunQueryDsl};
use lib::time::DateTimeInterval;
use serde::Deserialize;

#[cfg_attr(test, bon::builder)]
#[derive(Clone, Copy, PartialEq, Eq, Default, Deserialize)]
pub struct RoomQuery {
    pub available: Option<DateTimeInterval>,
    pub hotel_id: Option<i32>,
    pub min_beds: Option<i32>,
    pub min_size: Option<i32>,
}

impl Room {
    pub async fn query_rooms(
        RoomQuery {
            available,
            hotel_id,
            min_beds,
            min_size,
        }: RoomQuery,
        conn: &mut AsyncPgConnection,
    ) -> Result<Vec<Room>, AppError> {
        let mut table = room::table.left_outer_join(reservation::table).into_boxed();
        if let Some(DateTimeInterval { start, end }) = available {
            table = table.filter(
                reservation::id
                    .is_null()
                    .or(reservation::end.le(start))
                    .or(reservation::start.ge(end)),
            )
        }
        if let Some(id) = hotel_id {
            table = table.filter(room::hotel_id.eq(id));
        }
        if let Some(beds) = min_beds {
            table = table.filter(room::beds.ge(beds));
        }
        if let Some(size) = min_size {
            table = table.filter(room::size.ge(size));
        }
        table
            .select(room::all_columns)
            .get_results(conn)
            .await
            .map_err(Into::into)
    }
}

#[cfg(test)]
mod tests {
    use super::*;
    use crate::config;
    use crate::models::hotel::{CreateHotel, Hotel};
    use crate::models::reservation::{CreateReservation, Reservation};
    use crate::models::room::Room;
    use crate::models::user::{CreateUser, User};
    use diesel_async::AsyncPgConnection;
    use lib::diesel_crud_trait::DieselCrudCreate;
    use lib::test::diesel_pool::setup_test_transaction;
    use serde_json::json;

    #[rstest]
    #[case::all_none(RoomQuery::default(), &[1, 2, 3, 4])]
    #[case::min_beds_are_2(
        RoomQuery::builder()
            .min_beds(2)
            .build(),
        &[3, 4]
    )]
    #[case::min_beds_are_greater_than_all(
        RoomQuery::builder()
            .hotel_id(5)
            .build(),
        &[]
    )]
    #[case::min_size_are_5(
        RoomQuery::builder()
            .min_size(5)
            .build(),
        &[2, 4]
    )]
    #[case::min_size_is_greater_than_all(
        RoomQuery::builder()
            .min_size(10)
            .build(),
        &[]
    )]
    #[case::min_size_and_min_beds(
        RoomQuery::builder()
            .min_size(5)
            .min_beds(2)
            .build(),
        &[4]
    )]
    #[case::interval_when_no_reservations_in_interval(
        RoomQuery::builder()
            .available(DateTimeInterval {
                start: chrono::Utc::now().naive_utc() + chrono::Duration::days(10),
                end: chrono::Utc::now().naive_utc() + chrono::Duration::days(20)
            })
            .build(),
        &[1, 2, 3, 4]
    )]
    #[case::interval_when_one_reservation_in_interval(
        RoomQuery::builder()
            .available(DateTimeInterval {
                start: chrono::Utc::now().naive_utc(),
                end: chrono::Utc::now().naive_utc() + chrono::Duration::days(10)
            })
            .build(),
        &[2, 3, 4]
    )]
    #[case::all_filters(
        RoomQuery::builder()
            .available(DateTimeInterval {
                start: chrono::Utc::now().naive_utc(),
                end: chrono::Utc::now().naive_utc() + chrono::Duration::days(10)
            })
            .hotel_id(1)
            .min_beds(2)
            .min_size(1)
            .build(),
        &[3]
    )]
    #[tokio::test]
    async fn test_query_rooms(
        #[future] setup: Setup,
        #[case] query: RoomQuery,
        #[case] expected_ids: &[i32],
    ) {
        let Setup {
            mut conn, rooms, ..
        } = setup.await;
        let result = Room::query_rooms(query, &mut conn).await.unwrap();
        assert_eq!(
            result,
            rooms
                .into_iter()
                .filter(|room| expected_ids.contains(&room.id))
                .collect::<Vec<_>>()
        );
    }

    #[fixture]
    async fn setup() -> Setup {
        let mut conn = setup_test_transaction(config::DATABASE_URL).await.unwrap();
        let hotels = insert_hotels(&mut conn).await;
        let rooms = insert_rooms(&mut conn, &hotels).await;
        let _reservation = insert_reservation(&mut conn, rooms[0].id).await;
        Setup { conn, rooms }
    }

    struct Setup {
        conn: AsyncPgConnection,
        rooms: Vec<Room>,
    }

    async fn insert_hotels(conn: &mut AsyncPgConnection) -> Vec<Hotel> {
        let hotels = vec![
            CreateHotel::new(Some(1), "hotel1".into(), "hotel1".into()),
            CreateHotel::new(Some(2), "hotel2".into(), "hotel2".into()),
        ];
        Hotel::insert_many(&hotels, conn).await.unwrap()
    }

    async fn insert_rooms(conn: &mut AsyncPgConnection, hotels: &[Hotel]) -> Vec<Room> {
        let rooms: Vec<Room> = json!(
            [
                { "beds": 1, "size": 1 },
                { "beds": 1, "size": 5 },
                { "beds": 2, "size": 1 },
                { "beds": 2, "size": 5 }
            ]
        )
        .as_array()
        .unwrap()
        .iter()
        .enumerate()
        .map(|(index, room)| Room {
            hotel_id: hotels[index % 2].id,
            beds: room["beds"].as_i64().unwrap() as i32,
            size: room["size"].as_i64().unwrap() as i32,
            id: index as i32 + 1,
        })
        .collect();
        Room::insert_many(&rooms, conn).await.unwrap();
        rooms
    }
    async fn insert_reservation(conn: &mut AsyncPgConnection, room_id: i32) -> Reservation {
        let user = User::insert(
            User::hash_from_credentials(CreateUser::new("test@test.test", "test")).unwrap(),
            conn,
        )
        .await
        .unwrap();
        let now = chrono::Utc::now().naive_utc();
        let reservation =
            CreateReservation::new(room_id, now, now + chrono::Duration::days(5), user.email);
        Reservation::insert(reservation, conn).await.unwrap()
    }
}
