use crate::error::AppError;
use crate::models::session::Session;
use crate::GetConnection;
use axum::async_trait;
use derive_more::Constructor;
use lib::diesel_crud_trait::{DieselCrudCreate, DieselCrudDelete, DieselCrudRead};
use std::fmt::Debug;
use tower_sessions::session::{Id, Record};
use tower_sessions::{session_store, SessionStore};

#[derive(Clone, Constructor)]
pub struct SessionService<Pool>
where
    Pool: GetConnection,
{
    pool: Pool,
}

impl<Pool> Debug for SessionService<Pool>
where
    Pool: GetConnection,
{
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("SessionService")
            .field("pool", &self.pool.status())
            .finish()
    }
}

impl<Pool> SessionService<Pool>
where
    Pool: GetConnection,
{
    async fn create(&self, session: Session) -> Result<Session, AppError> {
        Session::insert(session, self.pool.get().await?.as_mut())
            .await
            .map_err(Into::into)
    }
    async fn read(&self, session_id: String) -> Result<Session, AppError> {
        Session::read(session_id, self.pool.get().await?.as_mut())
            .await
            .map_err(Into::into)
    }

    async fn delete(&self, session_id: String) -> Result<Session, AppError> {
        Session::delete(session_id, self.pool.get().await?.as_mut())
            .await
            .map_err(Into::into)
    }
}

#[async_trait]
impl<Pool> SessionStore for SessionService<Pool>
where
    Pool: GetConnection + 'static,
{
    async fn create(&self, session_record: &mut Record) -> session_store::Result<()> {
        self.save(session_record).await
    }

    async fn save(&self, session_record: &Record) -> session_store::Result<()> {
        let Ok(model) = Session::try_from(session_record.clone()) else {
            return Err(session_store::Error::Backend(
                "Failed to parse record to session model".to_string(),
            ));
        };

        Self::create(self, model).await.map_err(|error| {
            session_store::Error::Backend(format!("Failed to save session: {}", error))
        })?;
        Ok(())
    }

    async fn load(&self, session_id: &Id) -> session_store::Result<Option<Record>> {
        match self.read(session_id.0.to_string()).await {
            Ok(session) => match session.try_into_record() {
                Ok(record) => Ok(Some(record)),
                Err(error) => Err(session_store::Error::Decode(error.to_string())),
            },
            Err(AppError::NotFound) => Ok(None),
            Err(error) => Err(session_store::Error::Backend(format!(
                "Failed to load session: {error}",
            ))),
        }
    }

    /// Session-fixation cycling attempts to delete the session with the old session ID.
    /// If the session does not exist, it is considered a success.
    async fn delete(&self, session_id: &Id) -> session_store::Result<()> {
        match Self::delete(self, session_id.0.to_string()).await {
            Ok(_) | Err(AppError::NotFound) => Ok(()),
            Err(error) => Err(session_store::Error::Backend(format!(
                "Failed to delete session: {error}",
            ))),
        }
    }
}

#[cfg(test)]
mod tests {
    // TODO
}
