use crate::auth::hash_password;
use crate::error::{AppError, ResponseError};
use crate::models::user::{CreateUser, User, UserCredentials};
use crate::GetConnection;
use axum::async_trait;
use axum_login::{AuthnBackend, UserId};
use base64ct::{Base64, Encoding};
use derive_more::Constructor;
use lib::diesel_crud_trait::{DieselCrudCreate, DieselCrudRead};
use secrecy::ExposeSecret;

#[derive(Clone, Constructor)]
pub struct UserService<Pool>
where
    Pool: GetConnection,
{
    pool: Pool,
}

impl<Pool> UserService<Pool>
where
    Pool: GetConnection,
{
    pub async fn insert(&self, create: CreateUser) -> Result<User, ResponseError> {
        User::insert(
            User::hash_from_credentials(create)?,
            self.pool.get().await?.as_mut(),
        )
        .await
        .map_err(Into::into)
    }

    async fn read(&self, email: String) -> Result<User, AppError> {
        User::read(email, self.pool.get().await?.as_mut())
            .await
            .map_err(Into::into)
    }
}

#[async_trait]
impl<Pool> AuthnBackend for UserService<Pool>
where
    Pool: GetConnection,
{
    type User = User;
    type Credentials = UserCredentials;
    type Error = AppError;

    async fn authenticate(
        &self,
        UserCredentials { email, password }: Self::Credentials,
    ) -> Result<Option<Self::User>, Self::Error> {
        let user = self.read(email.clone()).await?;
        let password = password.expose_secret();
        let mut input_hash = [0; 32];
        hash_password(password, user.salt.as_bytes(), &mut input_hash);

        let mut user_hash = [0; 32];
        Base64::decode(&user.hash, &mut user_hash)?;
        if user.email == email && user_hash == input_hash {
            Ok(Some(user))
        } else {
            Ok(None)
        }
    }

    async fn get_user(&self, email: &UserId<Self>) -> Result<Option<Self::User>, Self::Error> {
        Ok(self.read(email.to_string()).await.ok())
    }
}

#[cfg(test)]
mod tests {
    // TODO
}
